# Cost Aware Retraining for Machine Learning

Code repository for the "Cost-Aware Retraining for Machine Learning" research paper.

## Environment and Dependencies

Install dependencies with 

```
conda env create -f env.yaml
```

## Datasets

Details for downloading the real-world datasets (Elec,Airlines,Covertype) are in [data/README.md](./data/README.md).

The synthetic datasets (CovCon, Circles and Gauss) are always generated, and the logic can be found in the `generate_data` method inside the respective dataset script found in [src/datasets/](./src/datasets/)

## Retraining Algorithms

The CARA variants, Oracle algorithm and other retraining algorithms are present in [src/algorithms/](./src/algorithms/)

The cost staleness computations for the CARA variants are present in [src/qdm.py](./src/qdm.py) where the `bayes_query_data_model_loss` is the 0/1 loss function and the `compute_qdm_diff` is the model staleness cost $\overline{\Psi}$ as described in the paper.  

## Baselines 

The drift detection baselines and incremental baselines are present in [src/baselines/](./src/baselines/)

The results for the baselines are gathered using the [src/baseline.py](./src/baseline.py) script.

## Config

We use [Hydra](https://hydra.cc) to maintain our experiment configurations. This [file](src/config/exp1.yaml) contains the configuration for the analysis experiment which collects all the results. Here, you can change the retraining algorithms and baselines evaluated. Similarly, there are configuration for the [datsets](src/config/dataset/) and [models](src/config/model/).

### Overrides 

You can override the defaults to run specific experiments. For example, to run with a SGDClassifier instead of the default RandomForest then simply add `model=LR_SGD` to the scripts.

Similarly, you can override the datasets, query distributions and other parameters from the command line or the notebook. 

## Running Experiments via scripts

There are four steps to collect all the results:
1. Create Cost Matrices: 
```python
python src/create_cost_matrices.py -cn exp1 dataset=covcon  dataset/queries=static_middle
```
2. Find maximum retraining cost for evaluation:
```python
python src/find_max_retraining_cost.py -cn exp1 dataset=covcon  dataset/queries=static_middle
```
3. Get running times for model and cost:
```python
python src/get_running_times.py -cn exp1 dataset=covcon  dataset/queries=static_middle
```
4. Perform offline and online analysis:
```python
python src/run_retrain_cost_analysis.py -cn exp1 dataset=covcon  dataset/queries=static_middle
```

### Results

The results by defaults are stored in the `results/` base directory based on the parameter from the [config](./src/config/exp1.yaml). The 

### Run Experiments

There is also a [script](retrain_analysis.sh) to perform the above step over multiple model seeds.
Modify the script with the correct number of CPUS to collect the results. See [Configuration](#config) for more details on how to configure the experiments.

### Running Experiments with Notebook

The above steps can be also performed interactively as seen in the [Testing Notebook](notebooks/testing.ipynb).

