# Imports
import numpy as np
import pandas as pd
from typing import Optional

def gather_retrains(models,datas,queries,baseline,retrain_costs,**kwargs):
    result = baseline.run(models,datas,queries,**kwargs)
    detector_time = result.get('detector_time',np.nan)
    _df = pd.DataFrame({"retrain_cost":retrain_costs})
    _df["retrains"] = [result["retrains"]]*len(_df)
    _df["num_retrains"] = [result["num_retrains"]]*len(_df)
    _df["parameters"] = [result["parameters"]]*len(_df)
    _df["detector_time"] = [detector_time]*len(_df)
    _df["algorithm"] = baseline.name 
    return _df

def get_inc_baseline_df(result:pd.DataFrame,inc_baseline_name:str,retrain_costs:np.ndarray,k_inc:Optional[float]=1.0):
    num_partial_fits = result["num_partial_fits"]
    total_staleness_cost = result["cost_fns_values"]["staleness_cost"].sum()
    total_time = result["running_time"].get("total",None)
    training_time = result["running_time"].get("training_time",None)
    detector_time = result["running_time"].get("detector_time",None)
    _df = pd.DataFrame({"retrain_cost":retrain_costs})
    _df["retrains"] = [result["retrains"]]*len(_df)
    _df["num_retrains"] = [result["num_retrains"]]*len(_df)
    _df["parameters"] = [result["parameters"]] * len(_df)
    _df["batch_staleness_cost"] = [list(result["cost_fns_values"]["staleness_cost"])]*len(_df)
    _df["total_staleness_cost"] = [total_staleness_cost]*len(_df)
    _df["strategy_cost"] = [kappa*num_partial_fits*k_inc + total_staleness_cost for kappa in _df.retrain_cost.values]
    _df["batch_prequential_acc"] = [list(result["cost_fns_values"]["prequential_acc"])]* len(_df)
    _df["prequential_acc"] = [result["cost_fns_values"]["prequential_acc"].mean()]* len(_df)
    _df["batch_prequential_query_acc"] = [list(result["cost_fns_values"]["prequential_query_acc"])]* len(_df)
    _df["prequential_query_acc"] = [result["cost_fns_values"]["prequential_query_acc"].mean()]* len(_df)
    _df["num_partial_fits"] = [num_partial_fits]* len(_df)
    _df["total_running_time"] = [total_time]* len(_df)
    _df["model_training_time"] = [training_time]* len(_df)
    _df["detector_time"] = [detector_time]* len(_df)
    _df["algorithm"] = inc_baseline_name 
    return _df

def gather_baselines(models,datas,queries,baselines,retrain_costs,**kwargs):
    dfs = []
    for baseline in baselines:
        dfs.append(gather_retrains(models,datas,queries,baseline,retrain_costs))
    return pd.concat(dfs).reset_index(drop=True)

def gather_incremental_baselines(cfg,datas,queries,incremental_baselines,retrain_costs,**kwargs):
    dfs = []
    for baseline in incremental_baselines:
        result = baseline.run(cfg,datas,queries)
        _df = get_inc_baseline_df(result,baseline.name,retrain_costs)
        dfs.append(_df)
    return pd.concat(dfs).reset_index(drop=True)