from pathlib import Path

import numpy as np
import pandas as pd
from sklearn.preprocessing import MinMaxScaler

proj_root = Path(__file__).parent.parent.parent.resolve()
from utils import get_online_offline_data_dict


def get_data(data_dir:Path,T:int=100):
    electricity_dir = data_dir/"elec"
    df = pd.read_csv(electricity_dir/"raw.csv")
    X = df.iloc[:,:-1].values
    y = df.iloc[:,-1].values

    scaler = MinMaxScaler()

    X_scaled = scaler.fit_transform(X)
    X_split = np.array_split(X_scaled,T)
    y_split = np.array_split(y,T)

    return X_split,y_split


def load_dicts(cfg):
    data_dir = proj_root/cfg.dataset.data.dir
    X_full, y_full = get_data(data_dir,cfg.dataset.num_batches)
    offline_datas, online_datas = get_online_offline_data_dict(
        X_full,
        y_full,
        offline_split=cfg.dataset.offline_split,
        use_ray=cfg.use_ray,
        queries_split = cfg.dataset.queries.queries_split
    )
    if cfg.dataset.queries.name == "data":
        return offline_datas, online_datas, offline_datas, online_datas
