import numpy as np

def vary_cost_old(C,retrain_costs=np.linspace(0,1,100)):
    T = C.shape[0]-1
    avg_costs = [] 
    strategies = []
    for cost in retrain_costs:
        dp,indices = dp_iterative(C,cost)
        avg_costs.append(dp[T]/(T+1))
        optimal_strategy = get_optimal_retrains(T,indices)
        strategies.append(optimal_strategy)
    return avg_costs,strategies


def compute_strategy_cost(strategy,cost):
    T = cost.shape[0]-1
    indices = np.zeros((T+1))
    # set the retrain locations
    indices[strategy] = strategy
    # forward fill the array
    np.maximum.accumulate(indices,axis=0, out=indices)
    return cost[indices.astype(int),np.arange(T+1)].sum()
    
def dp_recursive(t,C,retrain_cost:float=0.0):
    # print(f"{t=},{k=}")
    if t == 0:
        return C[0,0],0 
    elif t<0:
        return np.inf,-1
    
    prev_cost,prev_index = dp_recursive(t-1,C,retrain_cost)

    if (rt_cost:=C[t,t] + prev_cost +retrain_cost) < (kp_cost:=C[prev_index,t] + prev_cost ): #and not np.allclose(rt_cost,kp_cost):
        # print(f"{t=},{k=}",rt_cost, t)
        return rt_cost, t
    else:
        # print(f"{t=},{k=}",kp_cost, keep_index)
        return kp_cost,prev_index


def dp_iterative(C,retrain_cost:float=0.0):
    T = C.shape[0]-1
    dp = np.full(fill_value=np.nan,shape=(T+2))
    retrain_dp = np.full(fill_value=-2,shape=(T+2),dtype=int)

    dp[0]=C[0,0]
    retrain_dp[0]=0
    dp[-1] = np.inf

    for t in range(1,T+1):
        prev_cost = dp[t-1]
        prev_index = retrain_dp[t-1]
        retrain_loss = C[t,t]+retrain_cost+prev_cost
        not_retrain_cost = C[prev_index,t] + prev_cost
        if retrain_loss < not_retrain_cost :#and not np.allclose(retrain_cost,not_retrain_cost):
            dp[t]=retrain_loss
            retrain_dp[t]=t
        else:
            dp[t]=not_retrain_cost
            retrain_dp[t]=prev_index
    
    return dp,retrain_dp

def get_optimal_retrains(t,retrain_dp,verbose=False):
    retrains=[]
    while True:
        if t<0:
            break        
        index = retrain_dp[t]
        if verbose: print(f"Retrain at {index=}")
        retrains.append(index)
        if index == t:
            t-=1
        else:
            t = retrain_dp[index-1]
    return retrains

def get_optimal_retrains_recursive(t,C,retrain_cost:float=0.0):
    retrains=[]
    while True:
        _,index = dp_recursive(t,C,retrain_cost)
        if index<0:
            break        
        print(f"Retrain at {index=}")
        retrains.append(index)
        if index == t:
            t-=1
        else:
            _,t = dp_recursive(index-1,C,retrain_cost)
    return retrains



def fill_table(T,C,retrain_cost:float=0.0):
    """Fills the DP table with the retraining budget constraints

    Args:
        T (int): The number of timesteps
        C (np.ndarray): The cost table (T,T)
        retrain_cost (float, optional): The cost to perform a retrain. Defaults to 0.0.

    Returns:
        max_costs,indices: The DP table of C and retrain indices
    """
    # create DP table for max_C and retrian index
    min_costs = np.full(fill_value=np.nan,shape=(T+2,T+2))
    indices = np.full(fill_value=-2,shape=(T+2,T+2),dtype=int)
    # base cases
    min_costs[0,:]=C[0,0]
    min_costs[-1,:]=np.inf
    min_costs[:,-1]=np.inf
    indices[0,:]=0

    for t in range(1,T+1):
        for k in range(0,T+1):
            retrain_loss = C[t,t]+min_costs[t-1,k-1]+retrain_cost
            last_trained_index = indices[t-1,k]
            not_retrain_cost = C[last_trained_index,t] + min_costs[t-1,k]

            if retrain_loss < not_retrain_cost: #and not np.allclose(retrain_cost,not_retrain_cost):
                min_costs[t,k] = retrain_loss
                indices[t,k]=t
            else:
                min_costs[t,k] = not_retrain_cost
                indices[t,k] = last_trained_index

    return min_costs, indices

def get_retrains(t,k,indices,verbose=False):
    retrains=[]
    while True:
        if t<0:
            break        
        index = indices[t,k]
        if verbose: print(f"Retrain at {index=}")
        retrains.append(index)
        if index == t:
            t-=1
        else:
            t = indices[index-1,k]
        k-=1
    return retrains

def recusrive_min_costs(t,k,C,retrain_cost:float=0.0):
    # print(f"{t=},{k=}")
    if t == 0 and k>=0:
        return C[0,0],0 
    elif k<0 or t<0:
        return np.inf,-1

    retrain_loss, retrain_index = recusrive_min_costs(t-1,k-1,C,retrain_cost)
    keep_cost,keep_index = recusrive_min_costs(t-1,k,C,retrain_cost)

    if (rt_cost:=C[t,t] + retrain_loss + retrain_cost) < (kp_cost:=C[keep_index,t] + keep_cost): #and not np.allclose(rt_cost,kp_cost):
        # print(f"{t=},{k=}",rt_cost, t)
        return rt_cost, t
    else:
        # print(f"{t=},{k=}",kp_cost, keep_index)
        return kp_cost,keep_index

def get_recusrive_opt_solution(t,k,C,retrain_cost:float=0.0,verbose=False):
    retrains=[]
    while True:
        _,index = recusrive_min_costs(t,k,C,retrain_cost)
        if index<0:
            break        
        if verbose: print(f"Retrain at {index=}")
        retrains.append(index)
        if index == t:
            t-=1
        else:
            _,t = recusrive_min_costs(index-1,k,C)
        k-=1
    return retrains

