from pathlib import Path

import numpy as np


proj_root = Path(__file__).parent.parent.resolve()

import hydra
import json 

from analyse import find_max_retrain_cost
from utils import (load_cost_matrices)
from hydra.core.hydra_config import HydraConfig
import shutil
import re
import logging
log = logging.getLogger(__name__)

@hydra.main(version_base=None,config_path="config")
def run(cfg):
    results_dir = proj_root / cfg.results_dir
    # load all cost matrices
    max_cost = np.inf
    for file in results_dir.glob("**/*.npz"):
        log.info(f"Checking costs from {file}")
        # model_seed = re.search("model_seed_(\d+)",str(file)).group(1)
        cost_matrices = load_cost_matrices(file)
        max_retrain_cost = find_max_retrain_cost(cost_matrices["offline_C"])
        log.info(f"{max_retrain_cost=}")
        max_cost = min(max_retrain_cost,max_cost)
    if np.isinf(max_cost):
        raise ValueError(f"Max Cost if infinite check code.")
    log.info(f"Max Cost for evaluation is {max_cost}")
    with open(results_dir/"max_retrain_cost.txt","w+") as fp:
        fp.write(str(max_cost))
    
    log.info("Logging Configs and Logs to results directorty")
    log_dir = results_dir/"log"
    hydra_log_dir = HydraConfig.get().runtime.output_dir
    shutil.copytree(hydra_log_dir,str(log_dir),dirs_exist_ok=True)
    

if __name__ == "__main__":
    run()