from pathlib import Path

import numpy as np
import pandas as pd

proj_root = Path(__file__).parent.parent.resolve()

import hydra
import json 
from omegaconf import DictConfig
from analyse import find_max_retrain_cost
from utils import (import_obj)
import timeit
import re
from algorithms import periodic,cum_threshold,threshold,oracle,never_retrain,markov
from baselines import adwin,ddm,always_retrain
def get_model_training_time(cfg:DictConfig) -> float:
    orig_flag = cfg.use_ray
    # turn off ray
    cfg.use_ray=False
    load_fn = import_obj(cfg.dataset.load_fn)
    _, online_datas, _, _ = load_fn(cfg)
    data = online_datas[0]
    _X = data["X_train"]
    _y = data["y_train"]
    model_class = import_obj(cfg.model.class_fn)
    model = model_class(**cfg.model.kwargs)
    stmt = "model.fit(_X,_y)"
    timer = timeit.Timer(stmt=stmt,globals=locals())
    # find ideal number of iterations
    num,_ = timer.autorange()
    repeat = 3
    timings = timer.repeat(repeat,number=num)
    training_time = min(timings)/num
    # change flag back
    cfg.use_ray = orig_flag
    return training_time

def get_cost_time(cfg:DictConfig) -> float:
    orig_flag = cfg.use_ray
    # turn off ray
    cfg.use_ray=False
    load_fn = import_obj(cfg.dataset.load_fn)
    _, online_datas, _, online_queries = load_fn(cfg)
    model_data = online_datas[0]
    new_data = online_datas[1]
    queries = online_queries[1]
    _X = model_data["X_train"]
    _y = model_data["y_train"]
    model_class = import_obj(cfg.model.class_fn)
    model = model_class(**cfg.model.kwargs)
    model = model.fit(_X,_y)
    # load cost function
    cost_fn = import_obj(cfg.dataset.cost.fn)
    cost_fn_kwargs = dict(cfg.dataset.cost.kwargs)
    # if a custom qdm_fn was defined load it 
    if "qdm_fn" in cost_fn_kwargs:
        cost_fn_kwargs["qdm_fn"] = import_obj(cost_fn_kwargs["qdm_fn"])
    stmt = "cost_fn(model=model,model_data=model_data,new_data=new_data,query_data=queries,**cost_fn_kwargs)"
    timer = timeit.Timer(stmt=stmt,globals=locals())
    # find ideal number of iterations
    num,_ = timer.autorange()
    repeat = 3
    timings = timer.repeat(repeat,number=num)
    cost_time = min(timings)/num

    # change flag back
    cfg.use_ray = orig_flag
    return cost_time


def find_total_time(algorithm,num_batches,num_retrains,model_training_time,cost_time,**kwargs):
    if algorithm in [periodic.name,cum_threshold.name,threshold.name,markov.name]:
        return num_batches*cost_time + num_retrains*model_training_time
    elif algorithm in [adwin.name,ddm.name]:
        detector_time = kwargs["detector_time"]
        return detector_time + num_retrains*model_training_time
    elif algorithm == never_retrain.name:
        return 0
    elif algorithm == always_retrain.name:
        return num_retrains*model_training_time


def gather_total_time(row:pd.Series,num_batches,model_training_time,cost_time):
    detector_time = row.get("detector_time",None)
    return find_total_time(
        algorithm=row.algorithm,
        num_batches=num_batches,
        num_retrains=row.num_retrains,
        model_training_time=model_training_time,
        cost_time=cost_time,
        **{"detector_time":detector_time}
    )

@hydra.main(version_base=None,config_path="config")
def run(cfg):
    results_dir = proj_root / cfg.results_dir
    model_training_time = get_model_training_time(cfg)
    with open(results_dir/"model_training_time.txt","w+") as fp:
        fp.write(str(model_training_time))
    cost_time = get_cost_time(cfg)
    with open(results_dir/"cost_time.txt","w+") as fp:
        fp.write(str(cost_time))
    

if __name__ == "__main__":
    run()